@@ -1,6 +1,6 @@
 /* ssi - server-side-includes CGI program
 **
-** Copyright  1995 by Jef Poskanzer <jef@acme.com>.
+** Copyright  1995 by Jef Poskanzer <jef@mail.acme.com>.
 ** All rights reserved.
 **
 ** Redistribution and use in source and binary forms, with or without
@@ -62,8 +62,6 @@ internal_error( char* reason )
     char* title = "500 Internal Error";
 
     (void) printf( "\
-Content-type: text/html\n\
-\n\
 <HTML><HEAD><TITLE>%s</TITLE></HEAD>\n\
 <BODY><H2>%s</H2>\n\
 Something unusual went wrong during a server-side-includes request:\n\
@@ -80,8 +78,6 @@ not_found( char* filename )
     char* title = "404 Not Found";
 
     (void) printf( "\
-Content-type: text/html\n\
-\n\
 <HTML><HEAD><TITLE>%s</TITLE></HEAD>\n\
 <BODY><H2>%s</H2>\n\
 The requested server-side-includes filename, %s,\n\
@@ -714,6 +710,9 @@ main( int argc, char** argv )
     (void) strcpy( timefmt, "%a %b %e %T %Z %Y" );
     sizefmt = SF_BYTES;
 
+    /* The MIME type has to be text/html. */
+    (void) fputs( "Content-type: text/html\n\n", stdout );
+
     /* Get the name that we were run as. */
     script_name = getenv( "SCRIPT_NAME" );
     if ( script_name == (char*) 0 )
@@ -725,10 +724,7 @@ main( int argc, char** argv )
     /* Append the PATH_INFO, if any, to get the full URL. */
     path_info = getenv( "PATH_INFO" );
     if ( path_info == (char*) 0 )
-	{
-	internal_error( "Couldn't get PATH_INFO environment variable." );
-	exit( 1 );
-	}
+	path_info = "";
     url = (char*) malloc( strlen( script_name ) + strlen( path_info ) + 1 );
     if ( url == (char*) 0 )
 	{
@@ -759,9 +755,6 @@ main( int argc, char** argv )
 	exit( 1 );
 	}
 
-    /* The MIME type has to be text/html. */
-    (void) fputs( "Content-type: text/html\n\n", stdout );
-
     /* Read and handle the file. */
     read_file( path_info, path_translated, fp );
 
