@@ -1,16 +1,12 @@
 /* fdwatch.h - header file for fdwatch package
 **
-** This package abstracts the use of the select() and/or poll() system
-** calls.  The basic function of these calls is to watch a set of
-** file descriptors for activity.  select() originated in the BSD world,
+** This package abstracts the use of the select()/poll()/kqueue()
+** system calls.  The basic function of these calls is to watch a set
+** of file descriptors for activity.  select() originated in the BSD world,
 ** while poll() came from SysV land, and their interfaces are somewhat
 ** different.  fdwatch lets you write your code to a single interface,
 ** with the portability differences hidden inside the package.
 **
-** Furthermore, if your system implements both select() and poll(),
-** then fdwatch will use whichever call is most advantageous.  It may
-** actually switch back and forth at runtime, as the workload changes.
-**
 ** Usage is fairly simple.  Call fdwatch_get_nfiles() to initialize
 ** the package and find out how many fine descriptors are available.
 ** Then each time through your main loop, call fdwatch_clear(), then
@@ -23,7 +19,7 @@
 ** to save a little CPU time.
 **
 **
-** Copyright  1999 by Jef Poskanzer <jef@acme.com>.
+** Copyright  1999 by Jef Poskanzer <jef@mail.acme.com>.
 ** All rights reserved.
 **
 ** Redistribution and use in source and binary forms, with or without
@@ -63,11 +59,11 @@
 */
 extern int fdwatch_get_nfiles( void );
 
-/* Clear the fdwatch data structures. */
-extern void fdwatch_clear( void );
-
 /* Add a descriptor to the watch list.  rw is either FDW_READ or FDW_WRITE.  */
-extern void fdwatch_add_fd( int fd, int rw );
+extern void fdwatch_add_fd( int fd, void* client_data, int rw );
+
+/* Delete a descriptor from the watch list. */
+extern void fdwatch_del_fd( int fd );
 
 /* Do the watch.  Return value is the number of descriptors that are ready,
 ** or 0 if the timeout expired, or -1 on errors.  A timeout of INFTIM means
@@ -75,10 +71,15 @@ extern void fdwatch_add_fd( int fd, int rw );
 */
 extern int fdwatch( long timeout_msecs );
 
-/* Check if a descriptor was ready.  rw is either FDW_READ or FDW_WRITE.  */
-extern int fdwatch_check_fd( int fd, int rw );
+/* Check if a descriptor was ready. */
+extern int fdwatch_check_fd( int fd );
 
-/* Return usage stats on the fdwatch package. */
-extern void fdwatch_stats( long* nselectP, long* npollP );
+/* Get the client data for the next returned event.  Returns -1 when there
+** are no more events.
+*/
+extern void* fdwatch_get_next_client_data( void );
+
+/* Generate debugging statistics syslog message. */
+extern void fdwatch_logstats( long secs );
 
 #endif /* _FDWATCH_H_ */
