@@ -29,10 +29,10 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/time.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <stdio.h>
+#include <time.h>
 #include <fcntl.h>
 #include <syslog.h>
 #include <errno.h>
@@ -44,6 +44,10 @@
 #include "mmc.h"
 #include "libhttpd.h"
 
+#ifndef HAVE_INT64T
+typedef long long int64_t;
+#endif
+
 
 /* Defines. */
 #ifndef DEFAULT_EXPIRE_AGE
@@ -62,8 +66,12 @@
 #define INITIAL_HASH_SIZE (1 << 10)
 #endif
 
-#define max(a,b) ((a)>(b)?(a):(b))
-#define min(a,b) ((a)<(b)?(a):(b))
+#ifndef MAX
+#define MAX(a,b) ((a)>(b)?(a):(b))
+#endif
+#ifndef MIN
+#define MIN(a,b) ((a)<(b)?(a):(b))
+#endif
 
 
 /* The Map struct. */
@@ -317,11 +325,11 @@ mmc_cleanup( struct timeval* nowP )
     ** too many or too few files mapped.
     */
     if ( mapped_bytes > DESIRED_MAX_MAPPED_BYTES )
-	expire_age = max( ( expire_age * 2 ) / 3, DEFAULT_EXPIRE_AGE / 10 );
+	expire_age = MAX( ( expire_age * 2 ) / 3, DEFAULT_EXPIRE_AGE / 10 );
     else if ( map_count > DESIRED_MAX_MAPPED_FILES )
-	expire_age = max( ( expire_age * 2 ) / 3, DEFAULT_EXPIRE_AGE / 10 );
+	expire_age = MAX( ( expire_age * 2 ) / 3, DEFAULT_EXPIRE_AGE / 10 );
     else if ( map_count < DESIRED_MAX_MAPPED_FILES / 2 )
-	expire_age = min( ( expire_age * 5 ) / 4, DEFAULT_EXPIRE_AGE * 3 );
+	expire_age = MIN( ( expire_age * 5 ) / 4, DEFAULT_EXPIRE_AGE * 3 );
 
     /* Really free excess blocks on the free list. */
     while ( free_count > DESIRED_FREE_COUNT )
