@@ -64,6 +64,10 @@
 #define SHUT_WR 1
 #endif
 
+#ifndef HAVE_INT64T
+typedef long long int64_t;
+#endif
+
 
 static char* argv0;
 static int debug;
@@ -72,6 +76,7 @@ static char* dir;
 static char* data_dir;
 static int do_chroot, no_log, no_symlink_check, do_vhost, do_global_passwd;
 static char* cgi_pattern;
+static int cgi_limit;
 static char* url_pattern;
 static int no_empty_referers;
 static char* local_pattern;
@@ -95,11 +100,12 @@ typedef struct {
 static throttletab* throttles;
 static int numthrottles, maxthrottles;
 
-#define THROTTLE_NOLIMIT 1234567890L	/* default limit (if none) */
+#define THROTTLE_NOLIMIT -1
 
 
 typedef struct {
     int conn_state;
+    int next_free_connect;
     httpd_conn* hc;
     int tnums[MAXTHROTTLENUMS];         /* throttle indexes */
     int numtnums;
@@ -109,11 +115,11 @@ typedef struct {
     Timer* linger_timer;
     long wouldblock_delay;
     off_t bytes;
-    off_t bytes_to_send;
-    off_t bytes_sent;
+    off_t end_byte_index;
+    off_t next_byte_index;
     } connecttab;
 static connecttab* connects;
-static int numconnects, maxconnects;
+static int num_connects, max_connects, first_free_connect;
 static int httpd_conn_count;
 
 /* The connection states. */
@@ -131,8 +137,7 @@ long stats_connections;
 off_t stats_bytes;
 int stats_simultaneous;
 
-static int got_usr1;
-static int watchdog_flag;
+static int got_hup, got_usr1, watchdog_flag;
 
 
 /* Forwards. */
@@ -170,6 +175,8 @@ static void thttpd_logstats( long secs );
 static void
 handle_term( int sig )
     {
+    /* Don't need to set up the handler again, since it's a one-shot. */
+
     shut_down();
     syslog( LOG_NOTICE, "exiting due to signal %d", sig );
     closelog();
@@ -181,11 +188,13 @@ handle_term( int sig )
 static void
 handle_chld( int sig )
     {
+    const int oerrno = errno;
     pid_t pid;
     int status;
 
 #ifndef HAVE_SIGSET
-    (void) signal( SIGCHLD, handle_chld );	/* set up handler again */
+    /* Set up handler again. */
+    (void) signal( SIGCHLD, handle_chld );
 #endif /* ! HAVE_SIGSET */
 
     /* Reap defunct children until there aren't any more. */
@@ -200,7 +209,7 @@ handle_chld( int sig )
 	    break;
 	if ( (int) pid < 0 )
 	    {
-	    if ( errno == EINTR )	/* because of ptrace */
+	    if ( errno == EINTR || errno == EAGAIN )
 		continue;
 	    /* ECHILD shouldn't happen with the WNOHANG option,
 	    ** but with some kernels it does anyway.  Ignore it.
@@ -209,40 +218,41 @@ handle_chld( int sig )
 		syslog( LOG_ERR, "child wait - %m" );
 	    break;
 	    }
+	/* Decrement the CGI count.  Note that this is not accurate, since
+	** each CGI can involve two or even three child processes.
+	** Decrementing for each child means that when there is heavy CGI
+	** activity, the count will be lower than it should be, and therefore
+	** more CGIs will be allowed than should be.
+	*/
+	if ( hs != (httpd_server*) 0 )
+	    {
+	    --hs->cgi_count;
+	    if ( hs->cgi_count < 0 )
+		hs->cgi_count = 0;
+	    }
 	}
+
+    /* Restore previous errno. */
+    errno = oerrno;
     }
 
 
-/* SIGHUP says to re-open the log file.  Generally this doesn't work,
-** due to running chrooted.  What you have to do instead is send a SIGUSR1
-** to shut down cleanly, and then restart thttpd.  But the HUP code is here
-** anyway in case you absolutely have to try it.
-*/
+/* SIGHUP says to re-open the log file. */
 static void
 handle_hup( int sig )
     {
-    FILE* logfp;
+    const int oerrno = errno;
 
 #ifndef HAVE_SIGSET
-    (void) signal( SIGHUP, handle_hup );	/* set up handler again */
+    /* Set up handler again. */
+    (void) signal( SIGHUP, handle_hup );
 #endif /* ! HAVE_SIGSET */
 
-    got_usr1 = 0;
-    if ( no_log || hs == (httpd_server*) 0 )
-	return;
+    /* Just set a flag that we got the signal. */
+    got_hup = 1;
 
-    /* Re-open the log file. */
-    if ( logfile != (char*) 0 && strcmp( logfile, "-" ) != 0 )
-	{
-	logfp = fopen( logfile, "a" );
-	if ( logfp == (FILE*) 0 )
-	    {
-	    syslog( LOG_CRIT, "reopening %.80s - %m", logfile );
-	    return;
-	    }
-	(void) fcntl( fileno( logfp ), F_SETFD, 1 );
-	httpd_set_logfp( hs, logfp );
-	}
+    /* Restore previous errno. */
+    errno = oerrno;
     }
 
 
@@ -250,7 +260,9 @@ handle_hup( int sig )
 static void
 handle_usr1( int sig )
     {
-    if ( numconnects == 0 )
+    /* Don't need to set up the handler again, since it's a one-shot. */
+
+    if ( num_connects == 0 )
 	{
 	/* If there are no active connections we want to exit immediately
 	** here.  Not only is it faster, but without any connections the
@@ -261,8 +273,11 @@ handle_usr1( int sig )
 	closelog();
 	exit( 0 );
 	}
+
     /* Otherwise, just set a flag that we got the signal. */
     got_usr1 = 1;
+
+    /* Don't need to restore old errno, since we didn't do any syscalls. */
     }
 
 
@@ -270,11 +285,17 @@ handle_usr1( int sig )
 static void
 handle_usr2( int sig )
     {
+    const int oerrno = errno;
+
 #ifndef HAVE_SIGSET
-    (void) signal( SIGUSR2, handle_usr2 );	/* set up handler again */
+    /* Set up handler again. */
+    (void) signal( SIGUSR2, handle_usr2 );
 #endif /* ! HAVE_SIGSET */
 
     logstats( (struct timeval*) 0 );
+
+    /* Restore previous errno. */
+    errno = oerrno;
     }
 
 
@@ -282,6 +303,8 @@ handle_usr2( int sig )
 static void
 handle_alrm( int sig )
     {
+    const int oerrno = errno;
+
     /* If nothing has been happening */
     if ( ! watchdog_flag )
 	{
@@ -293,9 +316,38 @@ handle_alrm( int sig )
     watchdog_flag = 0;
 
 #ifndef HAVE_SIGSET
-    (void) signal( SIGALRM, handle_alrm );	/* set up handler again */
+    /* Set up handler again. */
+    (void) signal( SIGALRM, handle_alrm );
 #endif /* ! HAVE_SIGSET */
-    (void) alarm( OCCASIONAL_TIME * 3 );	/* set up alarm again */
+    /* Set up alarm again. */
+    (void) alarm( OCCASIONAL_TIME * 3 );
+
+    /* Restore previous errno. */
+    errno = oerrno;
+    }
+
+
+static void
+re_open_logfile( void )
+    {
+    FILE* logfp;
+
+    if ( no_log || hs == (httpd_server*) 0 )
+	return;
+
+    /* Re-open the log file. */
+    if ( logfile != (char*) 0 && strcmp( logfile, "-" ) != 0 )
+	{
+	syslog( LOG_NOTICE, "re-opening logfile" );
+	logfp = fopen( logfile, "a" );
+	if ( logfp == (FILE*) 0 )
+	    {
+	    syslog( LOG_CRIT, "re-opening %.80s - %m", logfile );
+	    return;
+	    }
+	(void) fcntl( fileno( logfp ), F_SETFD, 1 );
+	httpd_set_logfp( hs, logfp );
+	}
     }
 
 
@@ -304,12 +356,12 @@ main( int argc, char** argv )
     {
     char* cp;
     struct passwd* pwd;
-    uid_t uid;
-    gid_t gid;
-    char cwd[MAXPATHLEN];
+    uid_t uid = 32767;
+    gid_t gid = 32767;
+    char cwd[MAXPATHLEN+1];
     FILE* logfp;
     int num_ready;
-    int cnum, ridx;
+    int cnum;
     connecttab* c;
     httpd_conn* hc;
     httpd_sockaddr sa4;
@@ -348,6 +400,22 @@ main( int argc, char** argv )
     if ( throttlefile != (char*) 0 )
 	read_throttlefile( throttlefile );
 
+    /* If we're root and we're going to become another user, get the uid/gid
+    ** now.
+    */
+    if ( getuid() == 0 )
+	{
+	pwd = getpwnam( user );
+	if ( pwd == (struct passwd*) 0 )
+	    {
+	    syslog( LOG_CRIT, "unknown user - '%.80s'", user );
+	    (void) fprintf( stderr, "%s: unknown user - '%s'\n", argv0, user );
+	    exit( 1 );
+	    }
+	uid = pwd->pw_uid;
+	gid = pwd->pw_gid;
+	}
+
     /* Log file. */
     if ( logfile != (char*) 0 )
 	{
@@ -367,23 +435,28 @@ main( int argc, char** argv )
 		perror( logfile );
 		exit( 1 );
 		}
+	    if ( logfile[0] != '/' )
+		{
+		syslog( LOG_WARNING, "logfile is not an absolute path, you may not be able to re-open it" );
+		(void) fprintf( stderr, "%s: logfile is not an absolute path, you may not be able to re-open it\n", argv0 );
+		}
 	    (void) fcntl( fileno( logfp ), F_SETFD, 1 );
+	    if ( getuid() == 0 )
+		{
+		/* If we are root then we chown the log file to the user we'll
+		** be switching to.
+		*/
+		if ( fchown( fileno( logfp ), uid, gid ) < 0 )
+		    {
+		    syslog( LOG_WARNING, "fchown logfile - %m" );
+		    perror( "fchown logfile" );
+		    }
+		}
 	    }
 	}
     else
 	logfp = (FILE*) 0;
 
-    /* Figure out uid/gid from user. */
-    pwd = getpwnam( user );
-    if ( pwd == (struct passwd*) 0 )
-	{
-	syslog( LOG_CRIT, "unknown user - '%.80s'", user );
-	(void) fprintf( stderr, "%s: unknown user - '%s'\n", argv0, user );
-	exit( 1 );
-	}
-    uid = pwd->pw_uid;
-    gid = pwd->pw_gid;
-
     /* Switch directories if requested. */
     if ( dir != (char*) 0 )
 	{
@@ -480,6 +553,27 @@ main( int argc, char** argv )
 	    perror( "chroot" );
 	    exit( 1 );
 	    }
+	/* If we're logging and the logfile's pathname begins with the
+	** chroot tree's pathname, then elide the chroot pathname so
+	** that the logfile pathname still works from inside the chroot
+	** tree.
+	*/
+	if ( logfile != (char*) 0 && strcmp( logfile, "-" ) != 0 )
+	    {
+	    if ( strncmp( logfile, cwd, strlen( cwd ) ) == 0 )
+		{
+		(void) strcpy( logfile, &logfile[strlen( cwd ) - 1] );
+		/* (We already guaranteed that cwd ends with a slash, so leaving
+		** that slash in logfile makes it an absolute pathname within
+		** the chroot tree.)
+		*/
+		}
+	    else
+		{
+		syslog( LOG_WARNING, "logfile is not within the chroot tree, you will not be able to re-open it" );
+		(void) fprintf( stderr, "%s: logfile is not within the chroot tree, you will not be able to re-open it\n", argv0 );
+		}
+	    }
 	(void) strcpy( cwd, "/" );
 	/* Always chdir to / after a chroot. */
 	if ( chdir( cwd ) < 0 )
@@ -521,6 +615,7 @@ main( int argc, char** argv )
     (void) signal( SIGUSR2, handle_usr2 );
     (void) signal( SIGALRM, handle_alrm );
 #endif /* HAVE_SIGSET */
+    got_hup = 0;
     got_usr1 = 0;
     watchdog_flag = 0;
     (void) alarm( OCCASIONAL_TIME * 3 );
@@ -534,7 +629,7 @@ main( int argc, char** argv )
     hs = httpd_initialize(
 	hostname,
 	gotv4 ? &sa4 : (httpd_sockaddr*) 0, gotv6 ? &sa6 : (httpd_sockaddr*) 0,
-	port, cgi_pattern, charset, p3p, max_age, cwd, no_log, logfp,
+	port, cgi_pattern, cgi_limit, charset, p3p, max_age, cwd, no_log, logfp,
 	no_symlink_check, do_vhost, do_global_passwd, url_pattern,
 	local_pattern, no_empty_referers );
     if ( hs == (httpd_server*) 0 )
@@ -610,25 +705,28 @@ main( int argc, char** argv )
 	}
 
     /* Initialize our connections table. */
-    maxconnects = fdwatch_get_nfiles();
-    if ( maxconnects < 0 )
+    max_connects = fdwatch_get_nfiles();
+    if ( max_connects < 0 )
 	{
 	syslog( LOG_CRIT, "fdwatch initialization failure" );
 	exit( 1 );
 	}
-    maxconnects -= SPARE_FDS;
-    connects = NEW( connecttab, maxconnects );
+    max_connects -= SPARE_FDS;
+    connects = NEW( connecttab, max_connects );
     if ( connects == (connecttab*) 0 )
 	{
 	syslog( LOG_CRIT, "out of memory allocating a connecttab" );
 	exit( 1 );
 	}
-    for ( cnum = 0; cnum < maxconnects; ++cnum )
+    for ( cnum = 0; cnum < max_connects; ++cnum )
 	{
 	connects[cnum].conn_state = CNST_FREE;
+	connects[cnum].next_free_connect = cnum + 1;
 	connects[cnum].hc = (httpd_conn*) 0;
 	}
-    numconnects = 0;
+    connects[max_connects - 1].next_free_connect = -1;	/* end of link list */
+    first_free_connect = 0;
+    num_connects = 0;
     httpd_conn_count = 0;
 
     if ( hs != (httpd_server*) 0 )
@@ -641,18 +739,26 @@ main( int argc, char** argv )
 
     /* Main loop. */
     (void) gettimeofday( &tv, (struct timezone*) 0 );
-    while ( ( ! terminate ) || numconnects > 0 )
+    while ( ( ! terminate ) || num_connects > 0 )
 	{
+	/* Do we need to re-open the log file? */
+	if ( got_hup )
+	    {
+	    re_open_logfile();
+	    got_hup = 0;
+	    }
+
 	/* Do the fd watch. */
 	num_ready = fdwatch( tmr_mstimeout( &tv ) );
 	if ( num_ready < 0 )
 	    {
-	    if ( errno == EINTR )
+	    if ( errno == EINTR || errno == EAGAIN )
 		continue;       /* try again */
 	    syslog( LOG_ERR, "fdwatch - %m" );
 	    exit( 1 );
 	    }
 	(void) gettimeofday( &tv, (struct timezone*) 0 );
+
 	if ( num_ready == 0 )
 	    {
 	    /* No fd's are ready - run the timers. */
@@ -683,9 +789,8 @@ main( int argc, char** argv )
 	    }
 
 	/* Find the connections that need servicing. */
-	for ( ridx = 0; ridx < num_ready; ++ridx )
+	while ( ( c = (connecttab*) fdwatch_get_next_client_data() ) != (connecttab*) -1 )
 	    {
-	    c = (connecttab*) fdwatch_get_client_data( ridx );
 	    if ( c == (connecttab*) 0 )
 		continue;
 	    hc = c->hc;
@@ -707,13 +812,11 @@ main( int argc, char** argv )
 	    terminate = 1;
 	    if ( hs != (httpd_server*) 0 )
 		{
-		httpd_server* ths = hs;
-		hs = (httpd_server*) 0;
-		if ( ths->listen4_fd != -1 )
-		    fdwatch_del_fd( ths->listen4_fd );
-		if ( ths->listen6_fd != -1 )
-		    fdwatch_del_fd( ths->listen6_fd );
-		httpd_terminate( ths );
+		if ( hs->listen4_fd != -1 )
+		    fdwatch_del_fd( hs->listen4_fd );
+		if ( hs->listen6_fd != -1 )
+		    fdwatch_del_fd( hs->listen6_fd );
+		httpd_unlisten( hs );
 		}
 	    }
 	}
@@ -757,6 +860,11 @@ parse_args( int argc, char** argv )
 #else /* CGI_PATTERN */
     cgi_pattern = (char*) 0;
 #endif /* CGI_PATTERN */
+#ifdef CGI_LIMIT
+    cgi_limit = CGI_LIMIT;
+#else /* CGI_LIMIT */
+    cgi_limit = 0;
+#endif /* CGI_LIMIT */
     url_pattern = (char*) 0;
     no_empty_referers = 0;
     local_pattern = (char*) 0;
@@ -987,6 +1095,11 @@ read_config( char* filename )
 		value_required( name, value );
 		cgi_pattern = e_strdup( value );
 		}
+	    else if ( strcasecmp( name, "cgilimit" ) == 0 )
+		{
+		value_required( name, value );
+		cgi_limit = atoi( value );
+		}
 	    else if ( strcasecmp( name, "urlpat" ) == 0 )
 		{
 		value_required( name, value );
@@ -1354,7 +1467,7 @@ shut_down( void )
 
     (void) gettimeofday( &tv, (struct timezone*) 0 );
     logstats( &tv );
-    for ( cnum = 0; cnum < maxconnects; ++cnum )
+    for ( cnum = 0; cnum < max_connects; ++cnum )
 	{
 	if ( connects[cnum].conn_state != CNST_FREE )
 	    httpd_close_conn( connects[cnum].hc, &tv );
@@ -1387,7 +1500,6 @@ shut_down( void )
 static int
 handle_newconnect( struct timeval* tvP, int listen_fd )
     {
-    int cnum;
     connecttab* c;
     ClientData client_data;
 
@@ -1398,7 +1510,7 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
     for (;;)
 	{
 	/* Is there room in the connection table? */
-	if ( numconnects >= maxconnects )
+	if ( num_connects >= max_connects )
 	    {
 	    /* Out of connection slots.  Run the timers, then the
 	    ** existing connections, and maybe we'll free up a slot
@@ -1408,11 +1520,13 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
 	    tmr_run( tvP );
 	    return 0;
 	    }
-	/* Find a free connection entry. */
-	for ( cnum = 0; cnum < maxconnects; ++cnum )
-	    if ( connects[cnum].conn_state == CNST_FREE )
-		break;
-	c = &connects[cnum];
+	/* Get the first free connection entry off the free list. */
+	if ( first_free_connect == -1 || connects[first_free_connect].conn_state != CNST_FREE )
+	    {
+	    syslog( LOG_CRIT, "the connects free list is messed up" );
+	    exit( 1 );
+	    }
+	c = &connects[first_free_connect];
 	/* Make the httpd_conn if necessary. */
 	if ( c->hc == (httpd_conn*) 0 )
 	    {
@@ -1441,12 +1555,15 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
 	    return 1;
 	    }
 	c->conn_state = CNST_READING;
-	++numconnects;
+	/* Pop it off the free list. */
+	first_free_connect = c->next_free_connect;
+	c->next_free_connect = -1;
+	++num_connects;
 	client_data.p = c;
 	c->active_at = tvP->tv_sec;
 	c->wakeup_timer = (Timer*) 0;
 	c->linger_timer = (Timer*) 0;
-	c->bytes_sent = 0;
+	c->next_byte_index = 0;
 	c->numtnums = 0;
 
 	/* Set the connection file descriptor to no-delay mode. */
@@ -1455,8 +1572,8 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
 	fdwatch_add_fd( c->hc->conn_fd, c, FDW_READ );
 
 	++stats_connections;
-	if ( numconnects > stats_simultaneous )
-	    stats_simultaneous = numconnects;
+	if ( num_connects > stats_simultaneous )
+	    stats_simultaneous = num_connects;
 	}
     }
 
@@ -1493,17 +1610,16 @@ handle_read( connecttab* c, struct timeval* tvP )
 	}
     if ( sz < 0 )
 	{
-	/* Ignore EWOULDBLOCK errors.  At first glance you would think that
-	** connections returned by fdwatch as readable should never give an
-	** EWOULDBLOCK; however, this apparently can happen if a packet gets
-	** garbled.
+	/* Ignore EINTR and EAGAIN.  Also ignore EWOULDBLOCK.  At first glance
+	** you would think that connections returned by fdwatch as readable
+	** should never give an EWOULDBLOCK; however, this apparently can
+	** happen if a packet gets garbled.
 	*/
-	if ( errno != EWOULDBLOCK )
-	    {
-	    httpd_send_err(
-		hc, 400, httpd_err400title, "", httpd_err400form, "" );
-	    finish_connection( c, tvP );
-	    }
+	if ( errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK )
+	    return;
+	httpd_send_err(
+	    hc, 400, httpd_err400title, "", httpd_err400form, "" );
+	finish_connection( c, tvP );
 	return;
 	}
     hc->read_idx += sz;
@@ -1544,26 +1660,29 @@ handle_read( connecttab* c, struct timeval* tvP )
 	return;
 	}
 
-    /* Fill in bytes_to_send. */
+    /* Fill in end_byte_index. */
     if ( hc->got_range )
 	{
-	c->bytes_sent = hc->init_byte_loc;
-	c->bytes_to_send = hc->end_byte_loc + 1;
+	c->next_byte_index = hc->first_byte_index;
+	c->end_byte_index = hc->last_byte_index + 1;
 	}
     else if ( hc->bytes_to_send < 0 )
-	c->bytes_to_send = 0;
+	c->end_byte_index = 0;
     else
-	c->bytes_to_send = hc->bytes_to_send;
+	c->end_byte_index = hc->bytes_to_send;
 
     /* Check if it's already handled. */
     if ( hc->file_address == (char*) 0 )
 	{
 	/* No file address means someone else is handling it. */
-	c->bytes_sent = hc->bytes_sent;
+	int tind;
+	for ( tind = 0; tind < c->numtnums; ++tind )
+	    throttles[c->tnums[tind]].bytes_since_avg += hc->bytes_sent;
+	c->next_byte_index = hc->bytes_sent;
 	finish_connection( c, tvP );
 	return;
 	}
-    if ( c->bytes_sent >= c->bytes_to_send )
+    if ( c->next_byte_index >= c->end_byte_index )
 	{
 	/* There's nothing to send. */
 	finish_connection( c, tvP );
@@ -1584,18 +1703,25 @@ handle_read( connecttab* c, struct timeval* tvP )
 static void
 handle_send( connecttab* c, struct timeval* tvP )
     {
+    size_t max_bytes;
     int sz, coast;
     ClientData client_data;
     time_t elapsed;
     httpd_conn* hc = c->hc;
+    int tind;
+
+    if ( c->max_limit == THROTTLE_NOLIMIT )
+	max_bytes = 1000000000L;
+    else
+	max_bytes = c->max_limit / 4;	/* send at most 1/4 seconds worth */
 
     /* Do we need to write the headers first? */
     if ( hc->responselen == 0 )
 	{
 	/* No, just write the file. */
 	sz = write(
-	    hc->conn_fd, &(hc->file_address[c->bytes_sent]),
-	    MIN( c->bytes_to_send - c->bytes_sent, c->max_limit ) );
+	    hc->conn_fd, &(hc->file_address[c->next_byte_index]),
+	    MIN( c->end_byte_index - c->next_byte_index, max_bytes ) );
 	}
     else
 	{
@@ -1606,11 +1732,14 @@ handle_send( connecttab* c, struct timeval* tvP )
 
 	iv[0].iov_base = hc->response;
 	iv[0].iov_len = hc->responselen;
-	iv[1].iov_base = &(hc->file_address[c->bytes_sent]);
-	iv[1].iov_len = MIN( c->bytes_to_send - c->bytes_sent, c->max_limit );
+	iv[1].iov_base = &(hc->file_address[c->next_byte_index]);
+	iv[1].iov_len = MIN( c->end_byte_index - c->next_byte_index, max_bytes );
 	sz = writev( hc->conn_fd, iv, 2 );
 	}
 
+    if ( sz < 0 && errno == EINTR )
+	return;
+
     if ( sz == 0 ||
 	 ( sz < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ) ) )
 	{
@@ -1639,6 +1768,7 @@ handle_send( connecttab* c, struct timeval* tvP )
 	    }
 	return;
 	}
+
     if ( sz < 0 )
 	{
 	/* Something went wrong, close this connection.
@@ -1680,11 +1810,13 @@ handle_send( connecttab* c, struct timeval* tvP )
 	    }
 	}
     /* And update how much of the file we wrote. */
-    c->bytes_sent += sz;
+    c->next_byte_index += sz;
     c->hc->bytes_sent += sz;
+    for ( tind = 0; tind < c->numtnums; ++tind )
+	throttles[c->tnums[tind]].bytes_since_avg += sz;
 
     /* Are we done? */
-    if ( c->bytes_sent >= c->bytes_to_send )
+    if ( c->next_byte_index >= c->end_byte_index )
 	{
 	/* This connection is finished! */
 	finish_connection( c, tvP );
@@ -1699,21 +1831,22 @@ handle_send( connecttab* c, struct timeval* tvP )
     if ( c->max_limit != THROTTLE_NOLIMIT )
 	{
 	elapsed = tvP->tv_sec - c->started_at;
-	if ( elapsed == 0 || c->hc->bytes_sent / elapsed > c->max_limit )
+	if ( elapsed == 0 )
+	    elapsed = 1;	/* count at least one second */
+	if ( c->hc->bytes_sent / elapsed > c->max_limit )
 	    {
 	    c->conn_state = CNST_PAUSING;
 	    fdwatch_del_fd( hc->conn_fd );
-	    /* When should we send the next c->max_limit bytes
-	    ** to get back on schedule?  If less than a second
-	    ** (integer math rounding), use 1/8 second.
+	    /* How long should we wait to get back on schedule?  If less
+	    ** than a second (integer math rounding), use 1/2 second.
 	    */
-	    coast = ( c->hc->bytes_sent + c->max_limit ) / c->max_limit - elapsed;
+	    coast = c->hc->bytes_sent / c->max_limit - elapsed;
 	    client_data.p = c;
 	    if ( c->wakeup_timer != (Timer*) 0 )
 		syslog( LOG_ERR, "replacing non-null wakeup_timer!" );
 	    c->wakeup_timer = tmr_create(
 		tvP, wakeup_connection, client_data,
-		coast ? ( coast * 1000L ) : 125L, 0 );
+		coast > 0 ? ( coast * 1000L ) : 500L, 0 );
 	    if ( c->wakeup_timer == (Timer*) 0 )
 		{
 		syslog( LOG_CRIT, "tmr_create(wakeup_connection) failed" );
@@ -1728,13 +1861,15 @@ handle_send( connecttab* c, struct timeval* tvP )
 static void
 handle_linger( connecttab* c, struct timeval* tvP )
     {
-    char buf[1024];
+    char buf[4096];
     int r;
 
     /* In lingering-close mode we just read and ignore bytes.  An error
     ** or EOF ends things, otherwise we go until a timeout.
     */
     r = read( c->hc->conn_fd, buf, sizeof(buf) );
+    if ( r < 0 && ( errno == EINTR || errno == EAGAIN ) )
+	return;
     if ( r <= 0 )
 	really_clear_connection( c, tvP );
     }
@@ -1783,39 +1918,60 @@ check_throttles( connecttab* c )
 static void
 clear_throttles( connecttab* c, struct timeval* tvP )
     {
-    int i, tnum;
+    int tind;
 
-    for ( i = 0; i < c->numtnums; ++i )
-	{
-	tnum = c->tnums[i];
-	--throttles[tnum].num_sending;
-	throttles[tnum].bytes_since_avg += c->hc->bytes_sent;
-	}
+    for ( tind = 0; tind < c->numtnums; ++tind )
+	--throttles[c->tnums[tind]].num_sending;
     }
 
 
 static void
 update_throttles( ClientData client_data, struct timeval* nowP )
     {
-    int tnum;
+    int tnum, tind;
+    int cnum;
+    connecttab* c;
+    long l;
 
+    /* Update the average sending rate for each throttle.  This is only used
+    ** when new connections start up.
+    */
     for ( tnum = 0; tnum < numthrottles; ++tnum )
 	{
-	throttles[tnum].rate =
-	    ( 3 * throttles[tnum].rate +
-	      throttles[tnum].bytes_since_avg / THROTTLE_TIME ) / 4;
+	throttles[tnum].rate = ( 2 * throttles[tnum].rate + throttles[tnum].bytes_since_avg / THROTTLE_TIME ) / 3;
 	throttles[tnum].bytes_since_avg = 0;
 	/* Log a warning message if necessary. */
-	if ( throttles[tnum].rate > throttles[tnum].max_limit )
+	if ( throttles[tnum].rate > throttles[tnum].max_limit && throttles[tnum].num_sending != 0 )
 	    {
 	    if ( throttles[tnum].rate > throttles[tnum].max_limit * 2 )
-		syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld GREATLY exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].max_limit );
+		syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld greatly exceeding limit %ld; %d sending", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].max_limit, throttles[tnum].num_sending );
 	    else
-		syslog( LOG_INFO, "throttle #%d '%.80s' rate %ld exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].max_limit );
+		syslog( LOG_INFO, "throttle #%d '%.80s' rate %ld exceeding limit %ld; %d sending", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].max_limit, throttles[tnum].num_sending );
 	    }
-	if ( throttles[tnum].rate < throttles[tnum].min_limit )
+	if ( throttles[tnum].rate < throttles[tnum].min_limit && throttles[tnum].num_sending != 0 )
 	    {
-	    syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld lower than minimum %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].min_limit );
+	    syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld lower than minimum %ld; %d sending", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].min_limit, throttles[tnum].num_sending );
+	    }
+	}
+
+    /* Now update the sending rate on all the currently-sending connections,
+    ** redistributing it evenly.
+    */
+    for ( cnum = 0; cnum < max_connects; ++cnum )
+	{
+	c = &connects[cnum];
+	if ( c->conn_state == CNST_SENDING || c->conn_state == CNST_PAUSING )
+	    {
+	    c->max_limit = THROTTLE_NOLIMIT;
+	    for ( tind = 0; tind < c->numtnums; ++tind )
+		{
+		tnum = c->tnums[tind];
+		l = throttles[tnum].max_limit / throttles[tnum].num_sending;
+		if ( c->max_limit == THROTTLE_NOLIMIT )
+		    c->max_limit = l;
+		else
+		    c->max_limit = MIN( c->max_limit, l );
+		}
 	    }
 	}
     }
@@ -1863,8 +2019,9 @@ clear_connection( connecttab* c, struct timeval* tvP )
 	}
     if ( c->hc->should_linger )
 	{
+	if ( c->conn_state != CNST_PAUSING )
+	    fdwatch_del_fd( c->hc->conn_fd );
 	c->conn_state = CNST_LINGERING;
-	fdwatch_del_fd( c->hc->conn_fd );
 	shutdown( c->hc->conn_fd, SHUT_WR );
 	fdwatch_add_fd( c->hc->conn_fd, c, FDW_READ );
 	client_data.p = c;
@@ -1886,8 +2043,9 @@ clear_connection( connecttab* c, struct timeval* tvP )
 static void
 really_clear_connection( connecttab* c, struct timeval* tvP )
     {
-    stats_bytes += c->bytes_to_send;
-    fdwatch_del_fd( c->hc->conn_fd );
+    stats_bytes += c->hc->bytes_sent;
+    if ( c->conn_state != CNST_PAUSING )
+	fdwatch_del_fd( c->hc->conn_fd );
     httpd_close_conn( c->hc, tvP );
     clear_throttles( c, tvP );
     if ( c->linger_timer != (Timer*) 0 )
@@ -1896,7 +2054,9 @@ really_clear_connection( connecttab* c, struct timeval* tvP )
 	c->linger_timer = 0;
 	}
     c->conn_state = CNST_FREE;
-    --numconnects;
+    c->next_free_connect = first_free_connect;
+    first_free_connect = c - connects;	/* division by sizeof is implied */
+    --num_connects;
     }
 
 
@@ -1906,7 +2066,7 @@ idle( ClientData client_data, struct timeval* nowP )
     int cnum;
     connecttab* c;
 
-    for ( cnum = 0; cnum < maxconnects; ++cnum )
+    for ( cnum = 0; cnum < max_connects; ++cnum )
 	{
 	c = &connects[cnum];
 	switch ( c->conn_state )
@@ -2015,19 +2175,11 @@ static void
 thttpd_logstats( long secs )
     {
     if ( secs > 0 )
-#ifdef HAVE_INT64T
 	syslog( LOG_INFO,
 	    "  thttpd - %ld connections (%g/sec), %d max simultaneous, %lld bytes (%g/sec), %d httpd_conns allocated",
 	    stats_connections, (float) stats_connections / secs,
 	    stats_simultaneous, (int64_t) stats_bytes,
 	    (float) stats_bytes / secs, httpd_conn_count );
-#else /* HAVE_INT64T */
-	syslog( LOG_INFO,
-	    "  thttpd - %ld connections (%g/sec), %d max simultaneous, %ld bytes (%g/sec), %d httpd_conns allocated",
-	    stats_connections, (float) stats_connections / secs,
-	    stats_simultaneous, (long) stats_bytes,
-	    (float) stats_bytes / secs, httpd_conn_count );
-#endif /* HAVE_INT64T */
     stats_connections = 0;
     stats_bytes = 0;
     stats_simultaneous = 0;
