@@ -1,6 +1,6 @@
 /* timers.h - header file for timers package
 **
-** Copyright  1995,1998,1999,2000 by Jef Poskanzer <jef@acme.com>.
+** Copyright  1995,1998,1999,2000 by Jef Poskanzer <jef@mail.acme.com>.
 ** All rights reserved.
 **
 ** Redistribution and use in source and binary forms, with or without
@@ -32,7 +32,7 @@
 
 #ifndef INFTIM
 #define INFTIM -1
-#endif INFTIM
+#endif /* INFTIM */
 
 /* ClientData is a random value that tags along with a timer.  The client
 ** can use it for whatever, and it gets passed to the callback when the
@@ -103,7 +103,7 @@ extern void tmr_cleanup( void );
 /* Cancel all timers and free storage, usually in preparation for exitting. */
 extern void tmr_destroy( void );
 
-/* Return usage stats on the timer package. */
-extern void tmr_stats( int* activeP, int* freeP );
+/* Generate debugging statistics syslog message. */
+extern void tmr_logstats( long secs );
 
 #endif /* _TIMERS_H_ */
